
# from collections import deque
from pathlib import Path
import sys
# from string import ascii_lowercase

import pygame

import init_for_dev  # noqa
from auraboros import engine
from auraboros.utilities import AssetFilePath, draw_grid_background
from auraboros.gametext import TextSurfaceFactory
from auraboros.gamescene import Scene, SceneManager
from auraboros.gameinput import Keyboard
from auraboros.ui import GameMenuSystem, GameMenuUI, MsgWindow
from auraboros.animation import AnimationImage, SpriteSheet
from auraboros.particle import Emitter
from auraboros import global_

engine.init(caption="Test Animation System")

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")

textfactory = TextSurfaceFactory()
textfactory.register_font(
    "misaki_gothic",
    pygame.font.Font(AssetFilePath.font("misaki_gothic.ttf"), 16))


class TestAnimImg(AnimationImage):
    def __init__(self):
        super().__init__()
        self.sprite_sheet = SpriteSheet(AssetFilePath.img("testsprite.png"))
        self.anim_frames: list[pygame.surface.Surface] = [
            self.sprite_sheet.image_by_area(0, 0, 32, 32),
            self.sprite_sheet.image_by_area(0, 32, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*2, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*3, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*4, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*3, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*2, 32, 32),
            self.sprite_sheet.image_by_area(0, 32, 32, 32)]
        self.anim_interval = 500
        self.loop_count = 2


class DebugScene(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        textfactory.set_current_font("misaki_gothic")
        self.test_anim_img = TestAnimImg()
        self.keyboard["menu"] = Keyboard()
        self.keyboard.set_current_setup("menu")
        self.menusystem = GameMenuSystem()
        self.keyboard["menu"].register_keyaction(
            pygame.K_UP, 0, 122, self.menusystem.menu_cursor_up)
        self.keyboard["menu"].register_keyaction(
            pygame.K_DOWN, 0, 122, self.menusystem.menu_cursor_down)
        self.keyboard["menu"].register_keyaction(
            pygame.K_z, 0, 0, self.menusystem.do_selected_action)
        self.menusystem.add_menu_item(
            "play", self.run_emitter, text="Play")
        self.menusystem.add_menu_item(
            "stop", self.pause_emitter, text="STOP")
        self.menusystem.add_menu_item(
            "reset", self.reset_animation, text="RESET")
        self.menuui = GameMenuUI(self.menusystem, textfactory, "filled_box")
        self.menuui.padding = 4
        self.msgbox = MsgWindow(textfactory.font())
        self.msgbox.padding = 4
        self.msgbox.text = "Press 'Z'"
        self.msgbox2 = MsgWindow(textfactory.font())
        self.msgbox2.padding = 4
        self.msgbox3 = MsgWindow(textfactory.font())
        self.msgbox3.padding = 4
        self.msgbox4 = MsgWindow(textfactory.font())
        self.msgbox4.padding = 4
        self.msgbox5 = MsgWindow(textfactory.font())
        self.msgbox5.padding = 4
        self.msgbox6 = MsgWindow(textfactory.font())
        self.msgbox6.padding = 4
        self.testemitter = Emitter()
        self.testemitter.x = global_.w_size[0] // 2
        self.testemitter.y = global_.w_size[1] // 2

    def run_emitter(self):
        self.testemitter.let_emit()
        if self.testemitter.is_lifetime_end():
            self.testemitter.respawn()

    def pause_emitter(self):
        self.testemitter.let_pause()

    def reset_animation(self):
        self.test_anim_img.reset_animation()
        pass

    def update(self, dt):
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_UP)
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_DOWN)
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_z)
        self.menuui.set_pos_to_center()
        self.menusystem.update()
        # self.test_anim_img
        self.msgbox2.text = \
            f"lifetime:{self.testemitter.lifetime}"
        self.msgbox3.text = \
            f"anim_interval:{self.test_anim_img.anim_interval} milliseconds"
        self.msgbox4.text = \
            f"anim_frame_id:{self.test_anim_img.anim_frame_id}"
        self.msgbox5.text = \
            f"is_playing:{self.test_anim_img.is_playing}"
        self.msgbox6.text = \
            f"loop_counter:{self.test_anim_img.loop_counter}"
        self.msgbox2.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1]
        self.msgbox3.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1]
        self.msgbox4.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1] +\
            self.msgbox3.calculate_ultimate_size()[1]
        self.msgbox5.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1] +\
            self.msgbox3.calculate_ultimate_size()[1] +\
            self.msgbox5.calculate_ultimate_size()[1]
        self.msgbox6.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1] +\
            self.msgbox3.calculate_ultimate_size()[1] +\
            self.msgbox5.calculate_ultimate_size()[1] +\
            self.msgbox5.calculate_ultimate_size()[1]
        self.testemitter.update()

    def draw(self, screen):
        draw_grid_background(screen, 16, (78, 78, 78))
        # self.test_anim_img.draw(screen)
        screen.blit(
            self.test_anim_img.image,
            (global_.w_size[0]//2-self.test_anim_img.image.get_width()//2,
             self.menuui.pos[1]-self.test_anim_img.image.get_height()))
        self.menuui.draw(screen)
        self.msgbox.draw(screen)
        self.msgbox2.draw(screen)
        self.msgbox3.draw(screen)
        self.msgbox4.draw(screen)
        self.msgbox5.draw(screen)
        self.msgbox6.draw(screen)
        self.testemitter.draw(screen)


scene_manager = SceneManager()
scene_manager.push(DebugScene(scene_manager))

if __name__ == "__main__":
    engine.run(scene_manager=scene_manager, fps=60)



from collections import UserDict
from typing import Callable, TypedDict

import pygame


class Keyboard:
    def __init__(self):
        self.keyaction_dict: dict[int, KeyActionItem] = {}

    def __getitem__(self, key):
        return self.keyaction_dict[key]

    def event(self, event):
        if event.type == pygame.KEYDOWN:
            if self.keyaction_dict.get(event.key):
                self.keyaction_dict[event.key][
                    "is_pressed"] = True
                self.keyaction_dict[event.key][
                    "last_time"] = pygame.time.get_ticks()
        if event.type == pygame.KEYUP:
            if self.keyaction_dict.get(event.key):
                self.keyaction_dict[event.key]["is_pressed"] = False

    def release_all_of_keys(self):
        for key in self.keyaction_dict.keys():
            self.keyaction_dict[key]["is_pressed"] = False

    def do_action_by_keyinput(self, key_const, ignore_inregistered_key=False):
        if self.keyaction_dict.get(key_const) is None\
                and ignore_inregistered_key:
            return
        IS_PRESSED = self.keyaction_dict[key_const]["is_pressed"]
        DELAY = self.keyaction_dict[key_const]["delay"]
        INTERVAL = self.keyaction_dict[key_const]["interval"]
        last_time = self.keyaction_dict[key_const]["last_time"]
        current_time = pygame.time.get_ticks()
        do_keydown = False
        do_keyup = False
        if IS_PRESSED:
            if current_time - last_time >= DELAY:
                if not self.keyaction_dict[key_const]["_first_input_finished"]:
                    self.keyaction_dict[key_const][
                        "_first_input_finished"] = True
                    do_keydown = True
                elif current_time - last_time >= INTERVAL:
                    do_keydown = True
                    self.keyaction_dict[key_const]["last_time"] = current_time
        else:
            if self.keyaction_dict[key_const]["_first_input_finished"]:
                do_keyup = True
                self.keyaction_dict[key_const]["_first_input_finished"] = False
        if self.keyaction_dict[key_const]["keydown_deactivated"]:
            do_keydown = False
        if self.keyaction_dict[key_const]["keyup_deactivated"]:
            do_keyup = False
        if do_keydown:
            return self.keyaction_dict[key_const]["keydown"]()
        elif do_keyup:
            return self.keyaction_dict[key_const]["keyup"]()

    def register_keyaction(
            self,
            key_const,
            delay, interval,
            keydown: Callable = lambda: None, keyup: Callable = lambda: None):

        self.keyaction_dict[key_const] = KeyActionItem({
            "keydown": keydown, "keyup": keyup,
            "delay": delay,
            "interval": interval,
            "is_pressed": False,
            "last_time": 0,
            "_first_input_finished": False,
            "keydown_deactivated": False,
            "keyup_deactivated": False})

    def deactivate_keyup(self, key_const):
        self.keyaction_dict[key_const]["keyup_deactivated"] = True

    def activate_keyup(self, key_const):
        self.keyaction_dict[key_const]["keyup_deactivated"] = False

    def deactivate_keydown(self, key_const):
        self.keyaction_dict[key_const]["keydown_deactivated"] = True

    def activate_keydown(self, key_const):
        self.keyaction_dict[key_const]["keydown_deactivated"] = False


class KeyActionItem(TypedDict):
    keydown: Callable
    keyup: Callable
    delay: int
    interval: int
    is_pressed: bool
    last_time: int
    _first_input_finished: bool
    keydown_deactivated: bool
    keyup_deactivated: bool


class KeyboardSetupDict(UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, item: Keyboard):
        if isinstance(item, Keyboard):
            self.data[key] = item
        else:
            raise TypeError("The value must be Keyboard object.")

    def __getitem__(self, key) -> Keyboard:
        return self.data[key]


class KeyboardManager(KeyboardSetupDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_setup: Keyboard = None
        self.current_setup_key = None

    def set_current_setup(self, key):
        if self.current_setup is not None:
            self.current_setup.release_all_of_keys()
        self.current_setup = self.data[key]
        self.current_setup_key = key



import pygame

from src.auraboros.gameinput import Keyboard

keyboard = Keyboard()


def test_register_keyaction():
    delay = 0
    interval = 0
    keyboard.register_keyaction(pygame.K_z, delay, interval, lambda: 2)
    assert callable(keyboard[pygame.K_z]["keydown"])
    assert callable(keyboard[pygame.K_z]["keydown"])
    assert keyboard[pygame.K_z]["delay"] == delay
    assert keyboard[pygame.K_z]["interval"] == interval


def test_do_action_by_keyinput():
    pygame.init()
    testing = True
    pygame.event.post(
        pygame.event.Event(
            pygame.KEYDOWN,
            {"unicode": "z", "key": 122, "mod": 4096, "scancode": 29}))
    while testing:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                testing = False
            keyboard.event(event)
            # if event.type == pygame.KEYDOWN:
            #     print(event)
        assert keyboard.do_action_by_keyinput(pygame.K_z) == 2
        break



from collections import OrderedDict
from typing import Union

import pygame


class Schedule:
    """指定した時間間隔で関数を実行するためのスケジュール機能を提供するクラス
    _schedule_list = [func, interval, last_time, added_time, is_active]
    """

    _schedule_list: list[dict] = []

    @classmethod
    def add(cls, func, interval):
        """関数をスケジュールに追加する。
        Args:
            func (function): 定期的に呼び出す関数
            interval (int): 関数を呼び出す間隔(milliseconds)
        """
        schedule = OrderedDict()
        schedule["func"] = func
        schedule["interval"] = interval
        schedule["last_time"] = None
        schedule["added_time"] = pygame.time.get_ticks()
        schedule["is_active"] = False
        cls._schedule_list.append(schedule)
        # cls._schedule_list.append(
        #     [func, interval, pygame.time.get_ticks(),
        #      pygame.time.get_ticks(), False])

    @classmethod
    def execute(cls):
        """スケジュールに登録された関数を実行する"""
        current_time = pygame.time.get_ticks()

        # for i, (func, interval, last_time, added_time, is_active)
        # in enumerate(
        #         cls._schedule_list):
        #     if current_time - last_time >= interval:
        #         func()
        #         cls._schedule_list[i][2] = current_time

        for schedule in cls._schedule_list:
            if not schedule["is_active"]:
                continue
            if current_time - schedule["last_time"] >= schedule["interval"]:
                schedule["func"]()
                # cls._schedule_list[i] = [
                #     func, interval, current_time, added_time, True]
                schedule["last_time"] = current_time
                # cls._schedule_list[i][2] = current_time

    @classmethod
    def remove(cls, func):
        """スケジュールから関数を削除する"""
        cls._schedule_list = [
            schedule for schedule in cls._schedule_list
            if schedule["func"] != func]

    @classmethod
    def is_func_scheduled(cls, func) -> bool:
        """スケジュールに指定した関数が登録されているかどうかを判定する"""
        return any([func == scheduled_func
                   for scheduled_func, _, _, _, _ in cls._schedule_list])

    @classmethod
    def get_schedule(cls, func) -> Union[dict, None]:
        """指定した関数オブジェクトが登録されているスケジュールを取得する"""
        for schedule in cls._schedule_list:
            if schedule["func"] == func:
                return schedule
        return None

    @classmethod
    def activate_schedule(cls, func):
        """スケジュールに登録した関数のインターバル実行を開始する。"""
        schedule = cls.get_schedule(func)
        schedule["is_active"] = True
        schedule["last_time"] = pygame.time.get_ticks()

    @classmethod
    def _debug(cls):
        for schedule in cls._schedule_list:
            schedule = list(schedule.values())
            print("func", id(schedule[0]),
                  "interval", schedule[1],
                  "elapsed", schedule[2],
                  "added", schedule[3],
                  "is active?", schedule[4])



import math
import random

import pygame


class Particle:

    def __init__(self):
        self.x = 0
        self.y = 0
        self.vx = 1
        self.vy = 1
        self.lifetime = 2000
        self.size = 3
        self.color = (255, 255, 255)
        self._spawn_time = None
        # self.size

    def update(self):
        if self._spawn_time is None:
            self._spawn_time = pygame.time.get_ticks()
        _current_time = pygame.time.get_ticks()
        if not _current_time - self._spawn_time >= self.lifetime:
            self.x += self.vx
            self.y += self.vy

    def draw(self, surface):
        pygame.draw.circle(surface, self.color,
                           (int(self.x), int(self.y)), self.size)


class Emitter:
    def __init__(self):
        self.x = 0
        self.y = 0
        # self.rate = rate
        self.lifetime = 1000
        self.particles: list[Particle] = []
        self.how_many_emit = -1  # -1 means endless during lifetime.
        self.emitted_counter = 0
        self._spawn_time = None
        self.is_emitting = False
        self.is_pausing = False

    def let_emit(self):
        self.is_emitting = True
        self.is_pausing = False

    def let_pause(self):
        self.is_emitting = False
        self.is_pausing = True

    def is_lifetime_end(self):
        if self._spawn_time:
            _current_time = pygame.time.get_ticks()
            return _current_time - self._spawn_time >= self.lifetime
        else:
            return False

    def respawn(self):
        self._spawn_time = pygame.time.get_ticks()
        self.emitted_counter = 0

    def update(self):
        _current_time = pygame.time.get_ticks()
        if self.is_emitting:
            if self._spawn_time is None:
                self.respawn()
            if not _current_time - self._spawn_time >= self.lifetime:
                if self.emitted_counter < self.how_many_emit or \
                        self.how_many_emit < 0:
                    particle = Particle()
                    particle.x = self.x
                    particle.y = self.y
                    speed = 1
                    angle = random.uniform(0, 360)
                    particle.vx = speed * math.cos(math.radians(angle))
                    particle.vy = speed * math.sin(math.radians(angle))
                    self.particles.append(particle)
                    self.emitted_counter += 1
            else:
                self.is_emitting = False
        if not self.is_pausing:
            for particle in self.particles:
                particle.update()
                if _current_time - particle._spawn_time >= particle.lifetime:
                    self.particles.remove(particle)

    def draw(self, screen: pygame.surface.Surface):
        for particle in self.particles:
            particle.draw(screen)



import moderngl
import pygame


from .global_ import init  # noqa
from .gamescene import SceneManager
from .schedule import Schedule
from . import global_
from .shader import Shader2D

clock = pygame.time.Clock()


def surface_to_texture(ctx: moderngl.Context, surface: pygame.surface.Surface):
    texture = ctx.texture(surface.get_size(), 4)
    texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
    # texture.swizzle = "BGRA" if windows
    return texture


def run(scene_manager: SceneManager, fps=60):
    display_flags = pygame.display.get_surface().get_flags()
    if display_flags & (pygame.DOUBLEBUF | pygame.OPENGL):
        opengl_is_used = True
        shader2d = Shader2D()
    else:
        opengl_is_used = False

    running = True

    while running:
        dt = clock.tick(fps) / 1000
        Schedule.execute()
        # Schedule._debug()
        global_.screen.fill((0, 0, 0))
        if opengl_is_used:
            pass
        for event in pygame.event.get():
            running = scene_manager.event(event)
        scene_manager.update(dt)
        scene_manager.draw(global_.screen)
        pygame.transform.scale(global_.screen, global_.w_size_unscaled,
                               pygame.display.get_surface())
        if opengl_is_used:
            shader2d.register_surface_as_texture(
                global_.screen, "display_surface")
            shader2d.use_texture("display_surface", 0)
            shader2d.render()
            pygame.display.flip()
        else:
            pygame.display.update()

    pygame.quit()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'auraboros'
copyright = '2023, Eleven-junichi2'
author = 'Eleven-junichi2'
release = '0.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

sys.path.insert(0, os.path.abspath('../src'))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
]

autodoc_member_order = 'bysource'

templates_path = ['_templates']
exclude_patterns = []

language = 'ja'

locale_dirs = ['locale/']   # path is example but recommended.
gettext_compact = False     # optional.

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



from pathlib import Path
# from array import array
import sys

import pygame
# import moderngl

import init_for_dev  # noqa
from auraboros import engine
from auraboros.animation import AnimationImage, SpriteSheet, AnimationDict
from auraboros.entity import Entity
from auraboros.gameinput import Keyboard
from auraboros.gamescene import Scene, SceneManager
from auraboros.shader import Shader2D, VERTEX_DEFAULT
from auraboros.utilities import Arrow, AssetFilePath, draw_grid_background

engine.init(pixel_scale=1, set_mode_flags=pygame.DOUBLEBUF | pygame.OPENGL)

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")


class EntityIdle(AnimationImage):
    def __init__(self):
        super().__init__()
        self.sprite_sheet = SpriteSheet(AssetFilePath.img("testsprite.png"))
        self.anim_frames: list[pygame.surface.Surface] = [
            self.sprite_sheet.image_by_area(0, 0, 32, 32),
            self.sprite_sheet.image_by_area(0, 32, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*2, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*3, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*4, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*3, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*2, 32, 32),
            self.sprite_sheet.image_by_area(0, 32, 32, 32)]
        self.anim_interval = 75
        self.loop_count = -1


class TestEntity(Entity):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.animation = AnimationDict()
        self.animation["idle"] = EntityIdle()
        self.animation["idle"].seek(self.animation["idle"].frame_num//4)
        self.image = self.animation["idle"].image
        self.rect = self.image.get_rect()
        self.movement_speed = 2

    def update(self):
        self.move_by_arrow()
        if self.is_moving:
            self.animation["idle"].let_play()
            self.image = self.animation["idle"].image
        else:
            self.animation["idle"].let_stop()


class DebugScene(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.testentity = TestEntity()
        self.testentity.set_x_to_center_of_screen()
        self.testentity.set_y_to_center_of_screen()
        self.keyboard["player"] = Keyboard()
        self.keyboard["player"].register_keyaction(
            pygame.K_LEFT, 0, 0,
            lambda: self.testentity.set_move_direction(Arrow.LEFT),
            lambda: self.testentity.cancel_move_direction(Arrow.LEFT))
        self.keyboard["player"].register_keyaction(
            pygame.K_UP, 0, 0,
            lambda: self.testentity.set_move_direction(Arrow.UP),
            lambda: self.testentity.cancel_move_direction(Arrow.UP))
        self.keyboard["player"].register_keyaction(
            pygame.K_RIGHT, 0, 0,
            lambda: self.testentity.set_move_direction(Arrow.RIGHT),
            lambda: self.testentity.cancel_move_direction(Arrow.RIGHT))
        self.keyboard["player"].register_keyaction(
            pygame.K_DOWN, 0, 0,
            lambda: self.testentity.set_move_direction(Arrow.DOWN),
            lambda: self.testentity.cancel_move_direction(Arrow.DOWN))
        self.keyboard.set_current_setup("player")
        shader2d = Shader2D()
        with open(Path(__file__).parent / "vignette.frag", "r") as f:
            vignette_frag = f.read()
        shader2d.compile_and_register_program(
            VERTEX_DEFAULT,
            vignette_frag, "vignette")
        shader2d.set_uniform("vignette", "radius", 0.67)
        shader2d.set_uniform("vignette", "softness", 0.33)

    def update(self, dt):
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_LEFT)
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_UP)
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_RIGHT)
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_DOWN)
        self.testentity.update()

    def draw(self, screen):
        draw_grid_background(screen, 32, (178, 178, 178))
        self.testentity.draw(screen)


scene_manager = SceneManager()
scene_manager.push(DebugScene(scene_manager))

if __name__ == "__main__":
    engine.run(scene_manager=scene_manager)



from pygame.surface import Surface
from src.auraboros.animation import AnimationImage


def test_anim_frames():
    animation = AnimationImage()
    animation.anim_frames = [
        Surface((32, 32)),
        Surface((32, 32)),
        Surface((32, 32))
    ]
    assert len(animation.anim_frames) == 3


def test_anim_interval():
    animation = AnimationImage()
    animation.anim_interval = 500
    assert animation.anim_interval == 500


def test_is_all_loop_finished():
    animation = AnimationImage()
    animation.loop_count = 2
    animation.loop_counter = 2
    assert animation.is_all_loop_finished()


def test_let_play():
    animation = AnimationImage()
    animation.is_playing = False
    animation.loop_count = 1
    animation.loop_counter = 1
    animation.let_play()
    assert animation.is_playing
    assert animation.loop_counter == 0


def test_let_stop():
    animation = AnimationImage()
    animation.is_playing = True
    animation.let_stop()
    assert not animation.is_playing


def test_seek():
    animation = AnimationImage()
    animation.anim_frames = [
        Surface((32, 32)),
        Surface((32, 32)),
        Surface((32, 32))
    ]
    animation.seek(1)
    assert animation.anim_frame_id == 1


def test_reset_current_loop():
    animation = AnimationImage()
    animation.anim_frames = [
        Surface((32, 32)),
        Surface((32, 32)),
        Surface((32, 32))
    ]
    animation.anim_frame_id = 2
    animation.reset_current_loop()
    assert animation.anim_frame_id == 0


def test_reset_animation():
    animation = AnimationImage()
    animation.anim_frames = [
        Surface((32, 32)),
        Surface((32, 32)),
        Surface((32, 32))
    ]
    animation.anim_frame_id = 2
    animation.loop_count = 1
    animation.loop_counter = 1
    animation.reset_animation()
    assert animation.anim_frame_id == 0
    assert animation.loop_counter == 0


def test_update_animation():
    animation = AnimationImage()
    animation.anim_frames = [
        Surface((32, 32)),
        Surface((32, 32)),
        Surface((32, 32))
    ]
    animation.anim_interval = 500
    animation.is_playing = True
    animation.loop_count = 1
    animation.update_animation()
    assert animation.anim_frame_id == 1
    animation.update_animation()
    assert animation.anim_frame_id == 2
    animation.update_animation()
    assert animation.anim_frame_id == 0
    assert not animation.is_playing



import pygame

# from .animation import SpriteSheet


class TerrainTile:
    def __init__(self):
        self.image = pygame.surface.Surface((0, 0))
        self.rect = self.image.get_rect()


class Terrain:
    tilemap = []



from inspect import isclass
from typing import Any, MutableMapping

import pygame

from .schedule import Schedule


class AnimationImage:
    """アニメーションのある画像を設定・描写するためのクラス

    Attributes:
        anim_frame_id (int): 現在のフレームを示すインデックス
        anim_interval (int): アニメーションの更新間隔（ミリ秒）
        image (pygame.surface.Surface): 現在のフレームの画像
        is_playing (bool): アニメーションが再生中かどうかを示すフラグ
        loop_count (int): アニメーションのループ回数。-1で無限ループ指定
        loop_counter (int): 現在のループ回数
    """

    def __init__(self):
        self._anim_frames: list[pygame.surface.Surface] = [
            pygame.surface.Surface((0, 0)), ]
        self.anim_frame_id = 0
        self._anim_interval = 1
        self.image = self.anim_frames[self.anim_frame_id]
        self.is_playing = False
        self.loop_count = -1
        self.loop_counter = 0

    @property
    def anim_frames(self):
        return self._anim_frames

    @anim_frames.setter
    def anim_frames(self, value):
        self._anim_frames = value
        self.image = self.anim_frames[self.anim_frame_id]

    @property
    def frame_num(self):
        return len(self.anim_frames)

    @property
    def anim_interval(self):
        return self._anim_interval

    @anim_interval.setter
    def anim_interval(self, value):
        Schedule.remove(self.update_animation)
        self._anim_interval = value
        Schedule.add(self.update_animation, self.anim_interval)

    def is_all_loop_finished(self):
        return self.loop_count > 0 and self.loop_counter >= self.loop_count

    def let_play(self):
        if not self.is_playing:
            Schedule.activate_schedule(self.update_animation)
        self.is_playing = True
        if self.is_all_loop_finished():
            self.loop_counter = 0

    def let_stop(self):
        self.is_playing = False

    def seek(self, frame_id: int):
        self.anim_frame_id = frame_id
        self.image = self.anim_frames[self.anim_frame_id]

    def reset_current_loop(self):
        self.anim_frame_id = 0
        self.image = self._anim_frames[self.anim_frame_id]

    def reset_animation(self):
        self.anim_frame_id = 0
        self.image = self._anim_frames[self.anim_frame_id]
        self.loop_counter = 0

    def update_animation(self):
        if self.is_playing and (self.loop_counter < self.loop_count or
                                self.loop_count < 0):
            self.anim_frame_id = (self.anim_frame_id + 1) % len(
                self._anim_frames)
            self.image = self._anim_frames[self.anim_frame_id]
            if self.anim_frame_id == 0:
                self.loop_counter += 1
                if self.is_all_loop_finished():
                    self.is_playing = False


class AnimationFactory(MutableMapping):
    """
    Examples:
        class ExampleAnimation(AnimationImage):
            pass
        a = AnimationFactory()
        a["animation_a"] = ExampleAnimation
        animation = a["jump_animation"]
        animation.let_play_animation()
    """

    def __init__(self, *args, **kwargs):
        self.__dict__: dict[Any, AnimationImage]
        self.__dict__.update(*args, **kwargs)
        # self.anim_action_id = 0

    # def register(self, animation: AnimationImage):
        # self.__setitem__()

    def __getitem__(self, key) -> AnimationImage:
        return self.__dict__[key]()

    def __setitem__(self, key, value: AnimationImage):
        if isclass(value):
            self.__dict__[key] = value
        else:
            raise ValueError("The value must not be instance.")

    def __delitem__(self, key):
        del self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)


class AnimationDict(MutableMapping):
    """
    Examples:
        class ExampleAnimation(AnimationImage):
            pass
        a = AnimationFactory()
        a["animation_a"] = ExampleAnimation()
        animation = a["jump_animation"]
        animation.let_play_animation()
    """

    def __init__(self, *args, **kwargs):
        self.__dict__: dict[Any, AnimationImage]
        self.__dict__.update(*args, **kwargs)

    def __getitem__(self, key) -> AnimationImage:
        return self.__dict__[key]

    def __setitem__(self, key, value: AnimationImage):
        if not isclass(value):
            self.__dict__[key] = value
        else:
            raise ValueError("The value must be instance.")

    def __delitem__(self, key):
        del self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)


class SpriteSheet:
    def __init__(self, filename):
        self.image = pygame.image.load(filename)

    def image_by_area(self, x, y, width, height) -> pygame.surface.Surface:
        """"""
        image = pygame.Surface((width, height))
        image.blit(self.image, (0, 0), (x, y, width, height))
        image.set_colorkey((0, 0, 0))
        # image = pg.transform.scale(image, (width // 2, height // 2))
        return image
